#include <Homie.h>
#include <Adafruit_NeoPixel.h>
#include <ArduinoOTA.h>
#include "NeoPatterns.h"
#include "font.h"
#ifdef __AVR__
#include <avr/power.h>
#endif

#define PIN D1  //data pin for ws2812 (pixelprojektor @ ctdo: PIN 2)
#define NUMPIXELS 64

void StripComplete() {
  return;
}

NeoPatterns strip = NeoPatterns(NUMPIXELS, PIN, NEO_GRB + NEO_KHZ800, &StripComplete);

HomieNode homieNode("pixel", "commands");


bool onSetColor(const HomieRange& range, const String& value) {
  if (!range.isRange || range.index < 0 || range.index > 1) {
    return false;
  }
  switch (range.index) {
    case 0:
      strip.SetColor1(value.toInt());
      break;
    case 1:
      strip.SetColor2(value.toInt());
      break;
  }
  homieNode.setProperty("color_" + String(range.index)).send(value);
}

bool onSetPixel(const HomieRange& range, const String& value) {
  if (!range.isRange) {
    strip.None();
    strip.ColorSet(value.toInt());
    homieNode.setProperty("pixel").send(value);
    return true;
  }
  if (range.index < 0 || range.index > strip.numPixels() - 1) {
    return false;
  }
  strip.None();
  strip.setPixelColor(range.index, value.toInt());
  strip.show();
  homieNode.setProperty("pixel_" + String(range.index)).send(value);
}

bool onSetBrightness(const HomieRange& range, const String& value) {
  long brightness = value.toInt();
  if (brightness < 0 || brightness > 255) {
    return false;
  }
  strip.setBrightness(brightness);
  strip.show();
  homieNode.setProperty("brightness").send(value);
}

bool onSetEffect(const HomieRange& range, const String& value) {
  String effect = value;
  effect.toLowerCase();
  if (effect == "scanner") {
    strip.Scanner(strip.Color(255, 0, 0));
  }
  else if (effect == "randomscanner") {
    strip.Scanner(strip.Color(255, 0, 0), 40, true);
  }
  else if (effect == "larsonspiral") {
    strip.Scanner(strip.Color(255, 0, 0), 40, true, true);
  }
  else if (effect == "rainbowcycle") {
    strip.RainbowCycle(50);
  }
  else if (effect == "theaterchase") {
    strip.TheaterChase(strip.Color(255, 0, 0), strip.Color(0, 0, 255), 100);
  }
  else if (effect == "fade") {
    strip.Fade(strip.Color(255, 0, 0), strip.Color(0, 0, 255), 200, 100);
  }
  else if (effect == "randomfade") {
    strip.RandomFade();
  }
  else if (effect == "smooth") { //example: smooth|[wheelspeed]|[smoothing]|[strength]     wheelspeed=1-255, smoothing=0-100, strength=1-255
    strip.Smooth(16, 80, 50, 40);
  }
  else {
    // Test whether command with parameters was sent
    int sep = value.indexOf("|");
    String command = value.substring(0, sep);
    String parameters = value.substring(sep + 1);
    if (command.equals("fill"))
    {
      strip.ColorSetParameters(parameters);
    } else {
      strip.None();
    }
  }
  homieNode.setProperty("effect").send(value);
}

bool onSetClear(const HomieRange& range, const String& value) {
  strip.None();
  strip.clear();
  strip.show();
  homieNode.setProperty("clear").send(value);
}

bool onSetLength(const HomieRange& range, const String& value) {
  strip.None();
  strip.clear();
  strip.show();
  int newLength = value.toInt();
  if (newLength > 0) {
    strip.updateLength(newLength);
  }
  homieNode.setProperty("length").send(value);
}

void loopHandler() {
  strip.Update();

}

void setup() {
  Serial.begin(115200);

  Homie_setFirmware("pixelprojektor", "1.0.0");
  Homie.setLoopFunction(loopHandler);

  homieNode.advertiseRange("pixel", 0, NUMPIXELS - 1).settable(onSetPixel);
  homieNode.advertiseRange("color", 0, 1).settable(onSetColor);
  homieNode.advertise("brightness").settable(onSetBrightness);
  homieNode.advertise("effect").settable(onSetEffect);
  homieNode.advertise("clear").settable(onSetClear);
  homieNode.advertise("length").settable(onSetLength);


  Homie.setup();
  
  strip.begin();
  strip.clear();
  strip.setBrightness(64);
  strip.show();


  ArduinoOTA.setHostname("pixelprojektor");
  ArduinoOTA.onStart([]() {
    strip.clear();
  });
  ArduinoOTA.onEnd([]() {
    strip.clear();
  });
  ArduinoOTA.onProgress([](unsigned int progress, unsigned int total) {
    strip.setPixelColor(progress / (total / NUMPIXELS), strip.Color(100, 0, 0));
    strip.show();
  });
  ArduinoOTA.begin();
}

void loop() {
  Homie.loop();
  ArduinoOTA.handle();
}

// Diese Effekte müssen nach dem Umbau wieder vorhanden sein:
/*
  case EFFECT_SPIRAL:
    led_spiral();
    break;
  case EFFECT_RANDOMFADE:
    led_randomfade();
    break;
  case EFFECT_CHASE:
    led_chase();
    break;
*/



/************  Old stuff  ************/
/*
  int fadespeedmax = 5; //1 to 255
  int iconCountStart = 0; //for percentage calculation
  int iconCountdown = 0; //0=off
  uint8_t iconchar = 0; //last displayed char
  int32_t iconcolor = 0; //last icon color

  long lastMillis = 0;
  long fpsdelay = 1000 / FPS;


  uint32_t wheel(byte WheelPos) {
  WheelPos = 255 - WheelPos;
  if (WheelPos < 85) {
    return strip.Color(255 - WheelPos * 3, 0, WheelPos * 3);
  }
  if (WheelPos < 170) {
    WheelPos -= 85;
    return strip.Color(0, WheelPos * 3, 255 - WheelPos * 3);
  }
  WheelPos -= 170;
  return strip.Color(WheelPos * 3, 255 - WheelPos * 3, 0);
  }



  void led_random()
  {
  for (int i = 0; i < strip.numPixels(); i++) {
    strip.setPixelColor(i, wheel(random(0, 255)));
  }
  strip.show();
  }

  void led_radar()
  {
  // "Sweep" in cirles...
  // line(0,0,950*cos(radians(iAngle)),-950*sin(radians(iAngle)));
  }

  void led_larson()
  {
  int EyeSize = 5;
  uint32_t rgb[3] = {0};
  wheelPos++;
  if (wheelPos >= 256) {
    wheelPos = 0;
  }
  int Color1 = wheel(wheelPos);

  for (int i = 0; i < strip.numPixels(); i++)
  {
    int realpos = numToSpiralPos(i);
    if (i == Index) // Scan Pixel to the right
    {
      strip.setPixelColor(realpos, Color1);
    }
    else if (i == ((strip.numPixels() - 1) * 2) - Index) // Scan Pixel to the left
    {
      strip.setPixelColor(realpos, Color1);
    }
    else // Fading tail
    {
      strip.setPixelColor(realpos, DimColor(strip.getPixelColor(realpos)));
    }
  }

  strip.show();
  Increment();
  }


  void led_spiral()
  {
  int every = 4;
  wheelPos++;
  int qp = Index % every;
  Index++;
  if (Index >= strip.numPixels() - 1) {
    Index = 0;
  }
  int q = Index % every;
  for (uint16_t i = 0; i < strip.numPixels(); i = i + every) {
    strip.setPixelColor(numToSpiralPos(i + q), wheel( (i + Index * 4) % 255)); //turn every "every" pixel on
  }
  for (uint16_t i = 0; i < strip.numPixels(); i = i + every) {
    strip.setPixelColor(numToSpiralPos(i + qp), 0);      //turn every "every" pixel off
  }
  strip.show();

  }

  void led_randomfade()
  {
  for (int i = 0; i < strip.numPixels(); i++) {
    pixelR_buffer[i] += random(0, random(0, fadespeedmax + 1) + 1); //use buffer red channel for color wheel
    strip.setPixelColor(i, wheel(pixelR_buffer[i]));
  }
  strip.show();
  }

  void led_icon(uint8_t fontchar, uint32_t iconcolor)
  {
  for (int i = 0; i < strip.numPixels(); i++) {
    uint64_t mask = 1LL << (uint64_t)i;

    if ( (font[fontchar]&mask) == 0) {
      strip.setPixelColor(numToPos(i), strip.Color(0, 0, 0)); //bit is 0 at pos i
    } else {

      float _brightness = 1.0 - ( (iconCountStart - iconCountdown) * 1.0 / iconCountStart );
      uint8_t _r = (uint8_t)(iconcolor >> 16);
      uint8_t _g = (uint8_t)(iconcolor >>  8);
      uint8_t _b = (uint8_t)iconcolor;
      strip.setPixelColor(numToPos(i), strip.Color(_r * _brightness, _g * _brightness, _b * _brightness)); //bit is 1 at pos i
    }

  }
  strip.show();
  }

  void set_randomBuffer()
  {
  for (int i = 0; i < strip.numPixels(); i++) {
    uint32_t c = wheel(random(0, 256));
    pixelR_buffer[i] = (uint8_t)(c >> 16);
    pixelG_buffer[i] = (uint8_t)(c >>  8);
    pixelB_buffer[i] = (uint8_t)c;
  }
  }




  bool effectHandler(const HomieRange& range, const String& value) {
  Homie.getLogger() << "-> " << value << endl;
  int sep = value.indexOf("|");

  String command = value.substring(0, sep);
  String parameters = value.substring(sep + 1);
  Homie.getLogger() << "command=" << command << " parameters=" << parameters << endl;

  if (command.equals("fill")) {
    effect = EFFECT_NONE;
    led_fill(parseColor(parameters));
  } else if (command.equals("off")) {
    effect = EFFECT_NONE;
    led_fill(strip.Color(0, 0, 0));
  } else if (command.equals("random")) {
    effect = EFFECT_NONE;
    led_random();
  } else if (command.equals("set")) { //example: set|37#ff003a
    effect = EFFECT_NONE;
    int x = parameters.substring(0, 1).toInt();
    int y = parameters.substring(1, 2).toInt();
    String cstr = parameters.substring(2, 9);
    strip.setPixelColor(xyToPos(x, y), parseColor(cstr));
    strip.show();
  } else if (command.equals("smooth")) { //example: smooth|[wheelspeed]|[smoothing]|[strength]     wheelspeed=1-255, smoothing=0-100, strength=1-255
    int sepparam = parameters.indexOf("|");
    int p1 = parameters.substring(0, sepparam).toInt();
    String parameters_part2 = parameters.substring(sepparam + 1);
    sepparam = parameters_part2.indexOf("|");
    int p2 = parameters_part2.substring(0, sepparam).toInt();
    int p3 = parameters_part2.substring(sepparam + 1).toInt();
    wheelSpeed = 16; //default, speed=+1 /frame
    if (p1 > 0) {
      wheelSpeed = p1;
    }
    smoothing = 80;
    if (p2 > 0) {
      smoothing = p2;
    }
    strength = 50;
    if (p3 > 0) {
      strength = p3;
    }
    Homie.getLogger() << "-- p1=" << p1 << " p2=" << p2 << " p3=" << p3 << endl;
    effect = EFFECT_SMOOTH;
    bufferClear();
    showBuffer();
    strip.show();
  } else if (command.equals("spiral")) {
    effect = EFFECT_SPIRAL;
    Index = 0;
    bufferClear();
    showBuffer();
    strip.show();
  } else if (command.equals("clearbuffer")) {
    bufferClear();
    showBuffer();
    strip.show();
  } else if (command.equals("randomfade")) { //example: randomfade|5
    int sepparam = parameters.indexOf("|");
    int p1 = parameters.substring(0, sepparam).toInt();
    fadespeedmax = 5;
    if (p1 > 0) {
      fadespeedmax = p1;
    }
    effect = EFFECT_RANDOMFADE;
    set_randomBuffer(); //initialize random
  } else if (command.equals("randombuffer")) {
    set_randomBuffer(); //set random
    showBuffer();
  } else if (command.equals("chase")) {
    effect = EFFECT_CHASE;
    bufferClear();
    showBuffer();
    strip.show();
  } else if (command.equals("radar")) {
    effect = EFFECT_RADAR;
    Index = 0;
    bufferClear();
    showBuffer();
    strip.show();
  } else if (command.equals("larson")) {
    effect = EFFECT_LARSON;
    Index = 0;
    bufferClear();
    showBuffer();
    strip.show();
  }
  return true;
  }

  bool pixelsHandler(const HomieRange& range, const String& value) {

  String remaining = value;
  int i = 0;
  effect = EFFECT_NONE;
  do {
    String current = remaining.substring(0, 7);
    Homie.getLogger() << i << ":" << current << endl;
    uint32_t currentcolor = parseColor(current);

    strip.setPixelColor(i, currentcolor);
    i++;

    remaining = remaining.substring(7);

  } while (remaining.length() > 2 && (i < strip.numPixels()));
  Homie.getLogger() << "  filling rest with black" << endl;
  while (i < strip.numPixels()) {
    strip.setPixelColor(numToPos(i), strip.Color(0, 0, 0));
    i++;
  }

  strip.show();

  return true;
  }

  bool iconHandler(const HomieRange& range, const String& value) {
  String _iconname = value;
  iconcolor = strip.Color(255, 255, 255); //default color
  if (value[0] == '#') { //color given
    iconcolor = parseColor(value.substring(0, 6));
    _iconname = value.substring(7); //  example: #ff00dc|A  (pipe will be ignored)
  }

  iconCountStart = FPS * 2;
  iconCountdown = iconCountStart;
  if (_iconname.length() == 1) { //only one character
    iconchar = value[0];
  } else {

  }

  strip.show();

  return true;
  }


*/
